// PluginInterface.h

#ifndef ZIP7_INC_PLUGIN_INTERFACE_H
#define ZIP7_INC_PLUGIN_INTERFACE_H

/*
#include "../../../../C/7zTypes.h"
#include "../../IDecl.h"

#define Z7_IFACE_CONSTR_PLUGIN(i, n) \
  Z7_DECL_IFACE_7ZIP(i, 0x0A, n) \
  { Z7_IFACE_COM7_PURE(i) };

#define Z7_IFACEM_IInitContextMenu(x) \
  x(InitContextMenu(const wchar_t *folder, const wchar_t * const *names, UInt32 numFiles)) \

Z7_IFACE_CONSTR_PLUGIN(IInitContextMenu, 0x00)

#define Z7_IFACEM_IPluginOptionsCallback(x) \
  x(GetProgramFolderPath(BSTR *value)) \
  x(GetProgramPath(BSTR *value)) \
  x(GetRegistryCUPath(BSTR *value)) \

Z7_IFACE_CONSTR_PLUGIN(IPluginOptionsCallback, 0x01)

#define Z7_IFACEM_IPluginOptions(x) \
  x(PluginOptions(HWND hWnd, IPluginOptionsCallback *callback)) \
  // x(GetFileExtensions(BSTR *extensions))

Z7_IFACE_CONSTR_PLUGIN(IPluginOptions, 0x02)
*/
#endif
