#define IDD_FOLDERS     2400
#define IDD_FOLDERS_2  12400

#define IDT_FOLDERS_WORKING_FOLDER      2401
#define IDR_FOLDERS_WORK_SYSTEM         2402
#define IDR_FOLDERS_WORK_CURRENT        2403
#define IDR_FOLDERS_WORK_SPECIFIED      2404
#define IDX_FOLDERS_WORK_FOR_REMOVABLE  2405
#define IDS_FOLDERS_SET_WORK_PATH_TITLE 2406

#define IDE_FOLDERS_WORK_PATH  100
#define IDB_FOLDERS_WORK_PATH  101
