// EnumFormatEtc.h

#ifndef ZIP7_INC_ENUMFORMATETC_H
#define ZIP7_INC_ENUMFORMATETC_H

#include "../../../Common/MyWindows.h"

HRESULT CreateEnumFormatEtc(UINT numFormats, const FORMATETC *formats, IEnumFORMATETC **enumFormat);

#endif
