#include "ListViewDialogRes.h"
#include "../../GuiCommon.rc"

#define xc 480
#define yc 320

IDD_LISTVIEW  DIALOG  0, 0, xs, ys  MY_MODAL_RESIZE_DIALOG_STYLE  MY_FONT
CAPTION "ListView"
{
  CONTROL  "List1", IDL_LISTVIEW, "SysListView32", LVS_REPORT | LVS_SHOWSELALWAYS |
            LVS_AUTOARRANGE | LVS_NOCOLUMNHEADER | WS_BORDER | WS_TABSTOP,
            m, m, xc, yc - bys - m
  OK_CANCEL
}
