// StringUtils.h

#ifndef ZIP7_INC_STRING_UTILS_H
#define ZIP7_INC_STRING_UTILS_H

#include "../../../Common/MyString.h"

void SplitStringToTwoStrings(const UString &src, UString &dest1, UString &dest2);
// UString JoinStrings(const UStringVector &srcStrings);

#endif
