#define IDD_BROWSE  95

#define IDL_BROWSE         100
#define IDT_BROWSE_FOLDER  101
#define IDE_BROWSE_PATH    102
#define IDC_BROWSE_FILTER  103

#define IDB_BROWSE_PARENT     110
#define IDB_BROWSE_CREATE_DIR 112
